@@ -1,8 +1,6 @@
 # Copyright (C) 2011-2012  Patrick Totzke <patricktotzke@gmail.com>
 # This file is released under the GNU GPL, version 3 or a later revision.
 # For further details see the COPYING file
-from __future__ import absolute_import
-
 import argparse
 import glob
 import logging
@@ -13,7 +11,7 @@ from ..settings.const import settings
 from ..helper import split_commandstring, string_decode
 
 
-class Command(object):
+class Command:
 
     """base class for commands"""
     repeatable = False
@@ -24,7 +22,7 @@ class Command(object):
         self.undoable = False
         self.help = self.__doc__
 
-    def apply(self, caller):
+    def apply(self, ui):
         """code that gets executed when this command is applied"""
         pass
 
@@ -35,11 +33,19 @@ class CommandCanceled(Exception):
     pass
 
 
+class SequenceCanceled(Exception):
+    """ Exception triggered when a command sequence has been cancelled by the
+    confirmsequence command
+    """
+    pass
+
+
 COMMANDS = {
     'search': {},
     'envelope': {},
     'bufferlist': {},
     'taglist': {},
+    'namedqueries': {},
     'thread': {},
     'global': {},
 }
@@ -91,8 +97,7 @@ class CommandArgumentParser(argparse.ArgumentParser):
         raise CommandParseError(message)
 
 
-class registerCommand(object):
-
+class registerCommand:
     """
     Decorator used to register a :class:`Command` as
     handler for command `name` in `mode` so that it
