@@ -1,19 +1,18 @@
 # Copyright (C) 2011-2012  Patrick Totzke <patricktotzke@gmail.com>
+# Copyright © 2018 Dylan Baker
 # This file is released under the GNU GPL, version 3 or a later revision.
 # For further details see the COPYING file
-from __future__ import absolute_import
-
 import argparse
 import datetime
 import email
+import email.policy
 import glob
 import logging
 import os
 import re
 import tempfile
 import textwrap
-
-from twisted.internet.defer import inlineCallbacks
+import traceback
 
 from . import Command, registerCommand
 from . import globals
@@ -23,12 +22,14 @@ from .. import commands
 from .. import crypto
 from ..account import SendingMailFailed, StoreMailError
 from ..db.errors import DatabaseError
-from ..errors import GPGProblem
-from ..helper import email_as_string
+from ..errors import GPGProblem, ConversionError
 from ..helper import string_decode
+from ..helper import call_cmd
+from ..helper import split_commandstring
 from ..settings.const import settings
 from ..settings.errors import NoMatchingAccount
 from ..utils import argparse as cargparse
+from ..utils.collections import OrderedSet
 
 
 MODE = 'envelope'
@@ -103,25 +104,27 @@ class RefineCommand(Command):
         Command.__init__(self, **kwargs)
         self.key = key
 
-    def apply(self, ui):
+    async def apply(self, ui):
         value = ui.current_buffer.envelope.get(self.key, '')
         cmdstring = 'set %s %s' % (self.key, value)
-        ui.apply_command(globals.PromptCommand(cmdstring))
+        await ui.apply_command(globals.PromptCommand(cmdstring))
 
 
 @registerCommand(MODE, 'save')
 class SaveCommand(Command):
     """save draft"""
-    def apply(self, ui):
+    async def apply(self, ui):
         envelope = ui.current_buffer.envelope
 
         # determine account to use
-        try:
-            account = settings.get_account_by_address(
-                envelope['From'], return_default=True)
-        except NoMatchingAccount:
-            ui.notify('no accounts set.', priority='error')
-            return
+        if envelope.account is None:
+            try:
+                envelope.account = settings.account_matching_address(
+                    envelope['From'], return_default=True)
+            except NoMatchingAccount:
+                ui.notify('no accounts set.', priority='error')
+                return
+        account = envelope.account
 
         if account.draft_box is None:
             msg = 'abort: Account for {} has no draft_box'
@@ -130,9 +133,8 @@ class SaveCommand(Command):
 
         mail = envelope.construct_mail()
         # store mail locally
-        # add Date header
-        mail['Date'] = email.utils.formatdate(localtime=True)
-        path = account.store_draft_mail(email_as_string(mail))
+        path = account.store_draft_mail(
+            mail.as_string(policy=email.policy.SMTP))
 
         msg = 'draft saved successfully'
 
@@ -142,15 +144,15 @@ class SaveCommand(Command):
             logging.debug('adding new mail to index')
             try:
                 ui.dbman.add_message(path, account.draft_tags + envelope.tags)
-                ui.apply_command(globals.FlushCommand())
-                ui.apply_command(commands.globals.BufferCloseCommand())
+                await ui.apply_command(globals.FlushCommand())
+                await ui.apply_command(commands.globals.BufferCloseCommand())
             except DatabaseError as e:
-                logging.error(e)
-                ui.notify('could not index message:\n%s' % e,
+                logging.error(str(e))
+                ui.notify('could not index message:\n%s' % str(e),
                           priority='error',
                           block=True)
         else:
-            ui.apply_command(commands.globals.BufferCloseCommand())
+            await ui.apply_command(commands.globals.BufferCloseCommand())
 
 
 @registerCommand(MODE, 'send')
@@ -169,8 +171,24 @@ class SendCommand(Command):
         self.envelope = envelope
         self.envelope_buffer = None
 
-    @inlineCallbacks
-    def apply(self, ui):
+    def _get_keys_addresses(self):
+        addresses = set()
+        for key in self.envelope.encrypt_keys.values():
+            for uid in key.uids:
+                addresses.add(uid.email)
+        return addresses
+
+    def _get_recipients_addresses(self):
+        tos = self.envelope.headers.get('To', [])
+        ccs = self.envelope.headers.get('Cc', [])
+        return {a for (_, a) in email.utils.getaddresses(tos + ccs)}
+
+    def _is_encrypted_to_all_recipients(self):
+        recipients_addresses = self._get_recipients_addresses()
+        keys_addresses = self._get_keys_addresses()
+        return recipients_addresses.issubset(keys_addresses)
+
+    async def apply(self, ui):
         if self.mail is None:
             if self.envelope is None:
                 # needed to close later
@@ -186,7 +204,7 @@ class SendCommand(Command):
                 warning = 'A modified version of ' * mod
                 warning += 'this message has been sent at %s.' % when
                 warning += ' Do you want to resend?'
-                if (yield ui.choice(warning, cancel='no',
+                if (await ui.choice(warning, cancel='no',
                                     msg_position='left')) == 'no':
                     return
 
@@ -203,17 +221,28 @@ class SendCommand(Command):
                 warning = textwrap.dedent("""\
                     Any BCC recipients will not be able to decrypt this
                     message. Do you want to send anyway?""").replace('\n', ' ')
-                if (yield ui.choice(warning, cancel='no',
+                if (await ui.choice(warning, cancel='no',
                                     msg_position='left')) == 'no':
                     return
 
-            clearme = ui.notify(u'constructing mail (GPG, attachments)\u2026',
+            # Check if an encrypted message is indeed encrypted to all its
+            # recipients.
+            if (self.envelope.encrypt
+                    and not self._is_encrypted_to_all_recipients()):
+                warning = textwrap.dedent("""\
+                    Message is not encrypted to all recipients. This means that
+                    not everyone will be able to decode and read this message.
+                    Do you want to send anyway?""").replace('\n', ' ')
+                if (await ui.choice(warning, cancel='no',
+                                    msg_position='left')) == 'no':
+                    return
+
+            clearme = ui.notify('constructing mail (GPG, attachments)…',
                                 timeout=-1)
 
             try:
                 self.mail = self.envelope.construct_mail()
-                self.mail['Date'] = email.utils.formatdate(localtime=True)
-                self.mail = email_as_string(self.mail)
+                self.mail = self.mail.as_string(policy=email.policy.SMTP)
             except GPGProblem as e:
                 ui.clear_notify([clearme])
                 ui.notify(str(e), priority='error')
@@ -224,19 +253,38 @@ class SendCommand(Command):
         # determine account to use for sending
         msg = self.mail
         if not isinstance(msg, email.message.Message):
-            msg = email.message_from_string(self.mail)
-        address = msg.get('From', '')
+            msg = email.message_from_string(
+                self.mail, policy=email.policy.SMTP)
+        address = msg.get('Resent-From', False) or msg.get('From', '')
         logging.debug("FROM: \"%s\"" % address)
         try:
-            account = settings.get_account_by_address(address,
-                                                      return_default=True)
+            account = settings.account_matching_address(address,
+                                                        return_default=True)
         except NoMatchingAccount:
             ui.notify('no accounts set', priority='error')
             return
         logging.debug("ACCOUNT: \"%s\"" % account.address)
 
-        # define callback
-        def afterwards(_):
+        # send out
+        clearme = ui.notify('sending..', timeout=-1)
+        if self.envelope is not None:
+            self.envelope.sending = True
+        try:
+            await account.send_mail(self.mail)
+        except SendingMailFailed as e:
+            if self.envelope is not None:
+                self.envelope.account = account
+                self.envelope.sending = False
+            ui.clear_notify([clearme])
+            logging.error(traceback.format_exc())
+            errmsg = 'failed to send: {}'.format(e)
+            ui.notify(errmsg, priority='error', block=True)
+        except StoreMailError as e:
+            ui.clear_notify([clearme])
+            logging.error(traceback.format_exc())
+            errmsg = 'could not store mail: {}'.format(e)
+            ui.notify(errmsg, priority='error', block=True)
+        else:
             initial_tags = []
             if self.envelope is not None:
                 self.envelope.sending = False
@@ -246,8 +294,13 @@ class SendCommand(Command):
             ui.clear_notify([clearme])
             if self.envelope_buffer is not None:
                 cmd = commands.globals.BufferCloseCommand(self.envelope_buffer)
-                ui.apply_command(cmd)
+                await ui.apply_command(cmd)
             ui.notify('mail sent successfully')
+            if self.envelope is not None:
+                if self.envelope.replied:
+                    self.envelope.replied.add_tags(account.replied_tags)
+                if self.envelope.passed:
+                    self.envelope.passed.add_tags(account.passed_tags)
 
             # store mail locally
             # This can raise StoreMailError
@@ -257,65 +310,51 @@ class SendCommand(Command):
             if path is not None:
                 logging.debug('adding new mail to index')
                 ui.dbman.add_message(path, account.sent_tags + initial_tags)
-                ui.apply_command(globals.FlushCommand())
-
-        # define errback
-        def send_errb(failure):
-            if self.envelope is not None:
-                self.envelope.sending = False
-            ui.clear_notify([clearme])
-            failure.trap(SendingMailFailed)
-            logging.error(failure.getTraceback())
-            errmsg = 'failed to send: %s' % failure.value
-            ui.notify(errmsg, priority='error', block=True)
-
-        def store_errb(failure):
-            failure.trap(StoreMailError)
-            logging.error(failure.getTraceback())
-            errmsg = 'could not store mail: %s' % failure.value
-            ui.notify(errmsg, priority='error', block=True)
-
-        # send out
-        clearme = ui.notify('sending..', timeout=-1)
-        if self.envelope is not None:
-            self.envelope.sending = True
-        d = account.send_mail(self.mail)
-        d.addCallback(afterwards)
-        d.addErrback(send_errb)
-        d.addErrback(store_errb)
+                await ui.apply_command(globals.FlushCommand())
 
 
 @registerCommand(MODE, 'edit', arguments=[
     (['--spawn'], {'action': cargparse.BooleanAction, 'default': None,
                    'help': 'spawn editor in new terminal'}),
     (['--refocus'], {'action': cargparse.BooleanAction, 'default': True,
-                     'help': 'refocus envelope after editing'})])
+                     'help': 'refocus envelope after editing'}),
+    (['--part'], {'help': 'which alternative to edit ("html" or "plaintext")',
+                  'choices': ['html', 'plaintext']}),
+])
 class EditCommand(Command):
     """edit mail"""
-    def __init__(self, envelope=None, spawn=None, refocus=True, **kwargs):
+    def __init__(self, envelope=None, spawn=None, refocus=True, part=None,
+                 **kwargs):
         """
         :param envelope: email to edit
         :type envelope: :class:`~alot.db.envelope.Envelope`
         :param spawn: force spawning of editor in a new terminal
         :type spawn: bool
         :param refocus: m
+        :param part: which alternative to edit
+        :type part: str
         """
         self.envelope = envelope
         self.openNew = (envelope is not None)
         self.force_spawn = spawn
         self.refocus = refocus
         self.edit_only_body = False
+        self.edit_part = settings.get('envelope_edit_default_alternative')
+        if part in ['html', 'plaintext']:
+            self.edit_part = part
+        logging.debug('edit_part: %s ' % self.edit_part)
+
         Command.__init__(self, **kwargs)
 
-    def apply(self, ui):
+    async def apply(self, ui):
         ebuffer = ui.current_buffer
         if not self.envelope:
-            self.envelope = ui.current_buffer.envelope
+            self.envelope = ebuffer.envelope
 
         # determine editable headers
-        edit_headers = set(settings.get('edit_headers_whitelist'))
+        edit_headers = OrderedSet(settings.get('edit_headers_whitelist'))
         if '*' in edit_headers:
-            edit_headers = set(self.envelope.headers)
+            edit_headers = OrderedSet(self.envelope.headers)
         blacklist = set(settings.get('edit_headers_blacklist'))
         if '*' in blacklist:
             blacklist = set(self.envelope.headers)
@@ -338,8 +377,10 @@ class EditCommand(Command):
             translate = settings.get_hook('post_edit_translate')
             if translate:
                 template = translate(template, ui=ui, dbm=ui.dbman)
+            logging.debug('target bodypart: %s' % self.edit_part)
             self.envelope.parse_template(template,
-                                         only_body=self.edit_only_body)
+                                         only_body=self.edit_only_body,
+                                         target_body=self.edit_part)
             if self.openNew:
                 ui.buffer_open(buffers.EnvelopeBuffer(ui, self.envelope))
             else:
@@ -347,7 +388,7 @@ class EditCommand(Command):
                 ebuffer.rebuild()
 
         # decode header
-        headertext = u''
+        headertext = ''
         for key in edit_headers:
             vlist = self.envelope.get_all(key)
             if not vlist:
@@ -363,8 +404,23 @@ class EditCommand(Command):
                 value = re.sub('[ \t\r\f\v]*\n[ \t\r\f\v]*', ' ', value)
                 headertext += '%s: %s\n' % (key, value)
 
+        # determine which part to edit
+        logging.debug('edit_part: %s ' % self.edit_part)
+        if self.edit_part is None:
+            # I can't access ebuffer in my constructor, hence the check here
+            if isinstance(ebuffer, buffers.EnvelopeBuffer):
+                if ebuffer.displaypart in ['html', 'src']:
+                    self.edit_part = 'html'
+                    logging.debug('displaypart: %s' % ebuffer.displaypart)
+        if self.edit_part == 'html':
+            bodytext = self.envelope.body_html
+            logging.debug('editing HTML source')
+        else:
+            self.edit_part = 'plaintext'
+            bodytext = self.envelope.body_txt
+            logging.debug('editing plaintext')
+
         # determine editable content
-        bodytext = self.envelope.body
         if headertext:
             content = '%s\n%s' % (headertext, bodytext)
             self.edit_only_body = False
@@ -393,7 +449,7 @@ class EditCommand(Command):
                                   spawn=self.force_spawn,
                                   thread=self.force_spawn,
                                   refocus=self.refocus)
-        ui.apply_command(cmd)
+        await ui.apply_command(cmd)
 
 
 @registerCommand(MODE, 'set', arguments=[
@@ -414,12 +470,18 @@ class SetCommand(Command):
         self.reset = not append
         Command.__init__(self, **kwargs)
 
-    def apply(self, ui):
+    async def apply(self, ui):
         envelope = ui.current_buffer.envelope
         if self.reset:
             if self.key in envelope:
                 del envelope[self.key]
         envelope.add(self.key, self.value)
+        # FIXME: handle BCC as well
+        # Currently we don't handle bcc because it creates a side channel leak,
+        # as the key of the person BCC'd will be available to other recievers,
+        # defeating the purpose of BCCing them
+        if self.key.lower() in ['to', 'from', 'cc'] and envelope.encrypt:
+            await utils.update_keys(ui, envelope)
         ui.current_buffer.rebuild()
 
 
@@ -435,8 +497,14 @@ class UnsetCommand(Command):
         self.key = key
         Command.__init__(self, **kwargs)
 
-    def apply(self, ui):
+    async def apply(self, ui):
         del ui.current_buffer.envelope[self.key]
+        # FIXME: handle BCC as well
+        # Currently we don't handle bcc because it creates a side channel leak,
+        # as the key of the person BCC'd will be available to other recievers,
+        # defeating the purpose of BCCing them
+        if self.key.lower() in ['to', 'from', 'cc']:
+            await utils.update_keys(ui, ui.current_buffer.envelope)
         ui.current_buffer.rebuild()
 
 
@@ -501,13 +569,16 @@ class SignCommand(Command):
                     ui.notify(str(e), priority='error')
                     return
             else:
-                try:
-                    acc = settings.get_account_by_address(envelope['From'])
-                except NoMatchingAccount:
-                    envelope.sign = False
-                    ui.notify('Unable to find a matching account',
-                              priority='error')
-                    return
+                if envelope.account is None:
+                    try:
+                        envelope.account = settings.account_matching_address(
+                            envelope['From'])
+                    except NoMatchingAccount:
+                        envelope.sign = False
+                        ui.notify('Unable to find a matching account',
+                                  priority='error')
+                        return
+                acc = envelope.account
                 if not acc.gpg_key:
                     envelope.sign = False
                     msg = 'Account for {} has no gpg key'
@@ -561,8 +632,7 @@ class EncryptCommand(Command):
         self.trusted = trusted
         Command.__init__(self, **kwargs)
 
-    @inlineCallbacks
-    def apply(self, ui):
+    async def apply(self, ui):
         envelope = ui.current_buffer.envelope
         if self.action == 'rmencrypt':
             try:
@@ -587,7 +657,7 @@ class EncryptCommand(Command):
                     tmp_key = crypto.get_key(keyid)
                     envelope.encrypt_keys[tmp_key.fpr] = tmp_key
             else:
-                yield utils.set_encrypt(ui, envelope, signed_only=self.trusted)
+                await utils.update_keys(ui, envelope, signed_only=self.trusted)
         envelope.encrypt = encrypt
         if not envelope.encrypt:
             # This is an extra conditional as it can even happen if encrypt is
@@ -605,7 +675,7 @@ class EncryptCommand(Command):
 @registerCommand(
     MODE, 'retag', forced={'action': 'set'},
     arguments=[(['tags'], {'help': 'comma separated list of tags'})],
-    help='set message tags.',
+    help='set message tags',
 )
 @registerCommand(
     MODE, 'untag', forced={'action': 'remove'},
@@ -622,16 +692,16 @@ class TagCommand(Command):
     """manipulate message tags"""
     repeatable = True
 
-    def __init__(self, tags=u'', action='add', **kwargs):
+    def __init__(self, tags='', action='add', **kwargs):
         """
         :param tags: comma separated list of tagstrings to set
-        :type tags: unicode
+        :type tags: str
         :param action: adds tags if 'add', removes them if 'remove', adds tags
                        and removes all other if 'set' or toggle individually if
                        'toggle'
         :type action: str
         """
-        assert isinstance(tags, unicode), 'tags should be a unicode string'
+        assert isinstance(tags, str), 'tags should be a unicode string'
         self.tagsstring = tags
         self.action = action
         Command.__init__(self, **kwargs)
@@ -652,3 +722,95 @@ class TagCommand(Command):
         envelope.tags = sorted(new)
         # reload buffer
         ui.current_buffer.rebuild()
+
+
+@registerCommand(
+    MODE, 'html2txt', forced={'action': 'html2txt'},
+    arguments=[(['cmd'], {'nargs': argparse.REMAINDER,
+                          'help': 'converter command to use'})],
+    help='convert html to plaintext alternative',
+)
+@registerCommand(
+    MODE, 'txt2html', forced={'action': 'txt2html'},
+    arguments=[(['cmd'], {'nargs': argparse.REMAINDER,
+                          'help': 'converter command to use'})],
+    help='convert plaintext to html alternative',
+)
+class BodyConvertCommand(Command):
+    def __init__(self, action=None, cmd=None):
+        self.action = action
+        self.cmd = cmd  # this comes as a space separated list
+        Command.__init__(self)
+
+    def convert(self, cmdlist, inputstring):
+        logging.debug("converting using %s" % cmdlist)
+        resultstring, errmsg, retval = call_cmd(cmdlist,
+                                                stdin=inputstring)
+        if retval != 0:
+            msg = 'converter "%s" returned with ' % cmdlist
+            msg += 'return code %d' % retval
+            if errmsg:
+                msg += ':\n%s' % errmsg
+            raise ConversionError(msg)
+        logging.debug("resultstring is \n" + resultstring)
+        return resultstring
+
+    def apply(self, ui):
+        ebuffer = ui.current_buffer
+        envelope = ebuffer.envelope
+
+        if self.action is "txt2html":
+            fallbackcmd = settings.get('envelope_txt2html')
+            cmd = self.cmd or split_commandstring(fallbackcmd)
+            if cmd:
+                envelope.body_html = self.convert(cmd, envelope.body_txt)
+
+        elif self.action is "html2txt":
+            fallbackcmd = settings.get('envelope_html2txt')
+            cmd = self.cmd or split_commandstring(fallbackcmd)
+            if cmd:
+                envelope.body_txt = self.convert(cmd, envelope.body_html)
+
+        ui.current_buffer.rebuild()
+
+
+@registerCommand(
+    MODE, 'display', help='change which body alternative to display',
+    arguments=[(['part'], {'help': 'part to show'})])
+class ChangeDisplaymodeCommand(Command):
+
+    """change wich body alternative is shown"""
+
+    def __init__(self, part=None, **kwargs):
+        """
+        :param part: which part to show
+        :type indent: 'plaintext', 'src', or 'html'
+        """
+        self.part = part
+        Command.__init__(self, **kwargs)
+
+    async def apply(self, ui):
+        ebuffer = ui.current_buffer
+        envelope = ebuffer.envelope
+
+        # make sure that envelope has html part if requested here
+        if self.part in ['html', 'src'] and not envelope.body_html:
+            await ui.apply_command(BodyConvertCommand(action='txt2html'))
+
+        ui.current_buffer.set_displaypart(self.part)
+        ui.update()
+
+
+@registerCommand(
+    MODE, 'removehtml',
+    help='remove HTML alternative from the envelope',
+)
+class RemoveHtmlCommand(Command):
+    def apply(self, ui):
+        ebuffer = ui.current_buffer
+        envelope = ebuffer.envelope
+
+        envelope.body_html = None
+        ebuffer.displaypart = 'plaintext'
+        ebuffer.rebuild()
+        ui.update()
