@@ -18,15 +18,9 @@ from ..settings.const import settings
 
 charset.add_charset('utf-8', charset.QP, charset.QP, 'utf-8')
 
-MISSING_HTML_MSG = ("This message contains a text/html part that was not "
-                    "rendered due to a missing mailcap entry. "
-                    "Please refer to item 5 in our FAQ: "
-                    "http://alot.rtfd.io/en/latest/faq.html")
-
 
 @functools.total_ordering
-class Message(object):
-
+class Message:
     """
     a persistent notmuch message object.
     It it uses a :class:`~alot.db.DBManager` for cached manipulation
@@ -100,7 +94,7 @@ class Message(object):
         return NotImplemented
 
     def get_email(self):
-        """returns :class:`email.Message` for this message"""
+        """returns :class:`email.email.EmailMessage` for this message"""
         path = self.get_filename()
         warning = "Subject: Caution!\n"\
                   "Message file is no longer accessible:\n%s" % path
@@ -266,22 +260,10 @@ class Message(object):
                         self._attachments.append(Attachment(part))
         return self._attachments
 
-    def accumulate_body(self):
-        """
-        returns bodystring extracted from this mail
-        """
+    def get_body_text(self):
+        """ returns bodystring extracted from this mail """
         # TODO: allow toggle commands to decide which part is considered body
-        eml = self.get_email()
-        bodytext = extract_body(eml)
-
-        # check if extracted body is empty but msg contains html parts
-        if (not bodytext and
-           'text/html' in (part.get_content_type() for part in eml.walk())):
-            return MISSING_HTML_MSG
-        return bodytext
-
-    def get_text_content(self):
-        return extract_body(self.get_email(), types=['text/plain'])
+        return extract_body(self.get_email())
 
     def matches(self, querystring):
         """tests if this messages is in the resultset for `querystring`"""
