@@ -16,7 +16,7 @@ from .buffers import BufferlistBuffer
 from .buffers import SearchBuffer
 from .commands import globals
 from .commands import commandfactory
-from .commands import CommandCanceled
+from .commands import CommandCanceled, SequenceCanceled
 from .commands import CommandParseError
 from .helper import split_commandline
 from .helper import string_decode
@@ -34,7 +34,7 @@ async def periodic(callable_, period, *args, **kwargs):
         await asyncio.sleep(period)
 
 
-class UI(object):
+class UI:
     """
     This class integrates all components of alot and offers
     methods for user interaction like :meth:`prompt`, :meth:`notify` etc.
@@ -145,6 +145,10 @@ class UI(object):
             self.notify(str(exception), priority='error')
         elif isinstance(exception, CommandCanceled):
             self.notify("operation cancelled", priority='error')
+        elif isinstance(exception, SequenceCanceled):
+            # This exception needs to trickle up to apply_commandline,
+            # then be handled explicitly.
+            raise exception
         else:
             logging.error(traceback.format_exc())
             msg = "{}\n(check the log for details)".format(exception)
@@ -267,7 +271,11 @@ class UI(object):
             for c in split_commandline(cmdline):
                 await apply_this_command(c)
         except Exception as e:
-            self._error_handler(e)
+            if isinstance(e, SequenceCanceled):
+                self.notify("sequence of operations cancelled",
+                            priority='error')
+            else:
+                self._error_handler(e)
 
     @staticmethod
     def _unhandled_input(key):
@@ -290,7 +298,7 @@ class UI(object):
         self._unlock_callback = afterwards
         self._locked = True
 
-    def prompt(self, prefix, text=u'', completer=None, tab=0, history=None):
+    def prompt(self, prefix, text='', completer=None, tab=0, history=None):
         """
         prompt for text input.
         This returns a :class:`asyncio.Future`, which will have a string value
@@ -362,7 +370,6 @@ class UI(object):
         shuts down user interface without cleaning up.
         Use a :class:`alot.commands.globals.ExitCommand` for a clean shutdown.
         """
-        exit_msg = None
         try:
             loop = asyncio.get_event_loop()
             loop.stop()
@@ -670,9 +677,9 @@ class UI(object):
         info['pending_writes'] = len(self.dbman.writequeue)
         info['input_queue'] = ' '.join(self.input_queue)
 
-        lefttxt = righttxt = u''
+        lefttxt = righttxt = ''
         if cb is not None:
-            lefttxt, righttxt = settings.get(btype + '_statusbar', (u'', u''))
+            lefttxt, righttxt = settings.get(btype + '_statusbar', ('', ''))
             lefttxt = string_decode(lefttxt, 'UTF-8')
             lefttxt = lefttxt.format(**info)
             righttxt = string_decode(righttxt, 'UTF-8')
@@ -685,7 +692,7 @@ class UI(object):
         footerright = urwid.Text(righttxt, align='right')
         columns = urwid.Columns([
             footerleft,
-            ('fixed', len(righttxt), footerright)])
+            ('pack', footerright)])
         footer_att = settings.get_theming_attribute('global', 'footer')
         return urwid.AttrMap(columns, footer_att)
 
